unit ICQCrypt;
(***********************************************************)
(*   ICQCrypt, v. 01 (initial release)
(*   function IcqEncrypt encrypts packets according to
(*   V5(99b) protocol crypting algorithm.
(*
(*   Written by Coban(alex@ritlabs.com), special for using
(*   in ICQ Password Whacker app.
(**********)
interface
uses
  SysUtils;

  procedure IcqEncrypt(var pbyBuf: array of Byte; buf_len: Integer);
  procedure EnChar(var src: array of Byte; var dest: array of Char);
  procedure EnByte(var src: array of Char; var dest: array of Byte);

const
  table : array[0..255] of Byte =
  (
        $59, $60, $37, $6B, $65, $62, $46, $48,
        $53, $61, $4C, $59, $60, $57, $5B, $3D,
        $5E, $34, $6D, $36, $50, $3F, $6F, $67,
        $53, $61, $4C, $59, $40, $47, $63, $39,
        $50, $5F, $5F, $3F, $6F, $47, $43, $69,
        $48, $33, $31, $64, $35, $5A, $4A, $42,
        $56, $40, $67, $53, $41, $07, $6C, $49,
        $58, $3B, $4D, $46, $68, $43, $69, $48,
        $33, $31, $44, $65, $62, $46, $48, $53,
        $41, $07, $6C, $69, $48, $33, $51, $54,
        $5D, $4E, $6C, $49, $38, $4B, $55, $4A,
        $62, $46, $48, $33, $51, $34, $6D, $36,
        $50, $5F, $5F, $5F, $3F, $6F, $47, $63,
        $59, $40, $67, $33, $31, $64, $35, $5A,
        $6A, $52, $6E, $3C, $51, $34, $6D, $36,
        $50, $5F, $5F, $3F, $4F, $37, $4B, $35,
        $5A, $4A, $62, $66, $58, $3B, $4D, $66,
        $58, $5B, $5D, $4E, $6C, $49, $58, $3B,
        $4D, $66, $58, $3B, $4D, $46, $48, $53,
        $61, $4C, $59, $40, $67, $33, $31, $64,
        $55, $6A, $32, $3E, $44, $45, $52, $6E,
        $3C, $31, $64, $55, $6A, $52, $4E, $6C,
        $69, $48, $53, $61, $4C, $39, $30, $6F,
        $47, $63, $59, $60, $57, $5B, $3D, $3E,
        $64, $35, $3A, $3A, $5A, $6A, $52, $4E,
        $6C, $69, $48, $53, $61, $6C, $49, $58,
        $3B, $4D, $46, $68, $63, $39, $50, $5F,
        $5F, $3F, $6F, $67, $53, $41, $25, $41,
        $3C, $51, $54, $3D, $5E, $54, $5D, $4E,
        $4C, $39, $50, $5F, $5F, $5F, $3F, $6F,
        $47, $43, $69, $48, $33, $51, $54, $5D,
        $6E, $3C, $31, $64, $35, $5A, $00, $00
  );

implementation
  procedure EnChar(var src: array of Byte; var dest: array of Char);
  var
    i: Integer;
  begin
    FillChar(dest, SizeOf(Dest), 0);
    for i := Low(src) to High(src) do
      dest[i] := Chr(src[i]);
  end;

  procedure EnByte(var src: array of Char; var dest: array of Byte);
  var
    i: Integer;
  begin
    FillChar(dest, SizeOf(dest), 0);
    for i := Low(src) to High(src) do
      dest[i] := Ord(src[i]);
  end;

  function slab(len, pos: LongInt): LongInt;
  begin
    if len - pos >= 4 then
      Result := 4
    else
      Result := len - pos;
  end;


  procedure IcqEncrypt(var pbyBuf: array of Byte; buf_len: Integer);
  var
    number1: LongInt; //number1
    number2: LongInt; //number2
    cc: LongInt; //checkcode (number1 xor number2)
    a: array[0..5] of LongInt; //checkcode to insert tmp values
    r1: Integer; //random1
    r2: Integer; //random2
    size: Integer; //sizeof cc
    i: Integer;
    j: Integer;
    pos: Integer; //encryption position
    _slab: LongInt;
    tmpUint,
    LeftI,
    rightI, key : Longint;
    length_: Integer; //sizeof pbyBuf

  begin
    length_ := buf_len;

    number1 := pbyBuf[8];
    number1 := number1 shl 8; number1 := number1 or pbyBuf[4];
    number1 := number1 shl 8; number1 := number1 or pbyBuf[2];
    number1 := number1 shl 8; number1 := number1 or pbyBuf[6];

    if (length_ <= $14 + sizeof(cc)) then
      Exit;

    r1 := $18 + random(length_ - $18);
    r2 := random($ff);

    number2 := r1;
    number2 := number2 shl 8;
    number2 := number2 or pbyBuf[r1];
    number2 := number2 shl 8;
    number2 := number2 or r2;
    number2 := number2 shl 8;
    number2 := number2 or table[r2];
    number2 := number2 xor $ff00ff;

    cc := number1 xor number2;

    size := sizeof(cc);
    j := $14;
    for i := 0 to size - 1 do
    begin
      pbyBuf[j] := cc shr (i * 8);
      inc(j);
    end;

    key := length_ * $68656c6c + cc;
    pos := $a;
    _slab := SLAB(length_, pos);

    while _slab > 0 do
    begin
      tmpUint := 0;
      rightI := 0;
      LeftI := _slab - 1;
      while LeftI >= 0 do
      begin
        tmpUint := tmpUint shl 8; tmpUint := tmpUint or pbyBuf[pos + LeftI];
        dec(LeftI);
      end;
      tmpUint := tmpUint xor (key + table[pos and $FF]);
      while rightI < _slab do
      begin
        pbyBuf[pos] := tmpUint shr (rightI * 8);
        Inc(pos);
        Inc(rightI);
      end;
      _slab := SLAB(length_, pos);
    end;

    a[1] := cc and $0000001F; a[2] := cc and $03E003E0;
    a[3] := cc and $F8000400; a[4] := cc and $0000F800;
    a[5] := cc and $041F0000;

    a[1] := a[1] shl $0C; a[2] := a[2] shl $01;
    a[3] := a[3] shr $0A; a[4] := a[4] shl $10;
    a[5] := a[5] shr $0F;

    cc := a[1] + a[2] + a[3] + a[4] + a[5];

    // insert the scrambled checkcode at position 0x14
    size := sizeof(cc);
    j := $14;
    for i := 0 to size - 1 do
    begin
      pbyBuf[j] := cc shr (i * 8);
      inc(j);
    end;
  end;
end.